// Copyright (c) 2014, The Constrain project authors.
// Please see the AUTHORS file for details.
// All rights reserved. Use of this source code is governed by
// a BSD 2-Clause License that can be found in the LICENSE file.

import 'package:constrain/constraint.dart';
import 'package:matcher/matcher.dart';
import 'package:constrain/validator.dart';

class Address {
  String street;

  String toString() => 'Address[street: $street]';
}

isNotEmpty() => isNot(isEmpty);

@Ensure(_eitherOlderThan20OrHasAtLeastTwoAddresses)
class Person {
  @Ensure(isPositive)
  @Ensure(_isBetween10and90)
  int age;

  @Ensure(isNotNull)
  @Ensure(isNotEmpty)
  @Ensure(_allStreetsStartWith15)
  List<Address> addresses;

  static _allStreetsStartWith15() => everyElement(hasStreet(startsWith("15")));
  static _isBetween10and90() => allOf(greaterThanOrEqualTo(10), lessThanOrEqualTo(90));
  static _eitherOlderThan20OrHasAtLeastTwoAddresses() => 
      anyOf(hasAge(greaterThan(20)), 
          hasAddresses(hasLength(greaterThanOrEqualTo("two"))));

  String toString() => 'Person[age: $age, addressses: $addresses]';
}


main() {
  final Person person = new Person()
    ..age = 6
    ..addresses = [new Address()..street = "16 blah st"];

  Validator v = new Validator();
  final violations = v.validate(person);
  print(violations);
}

Matcher hasAge(matcher) => personMatcher("age", matcher, (Person p) => p.age);

Matcher hasAddresses(matcher) => personMatcher("addresses", matcher, 
    (Person p) => p.addresses);

Matcher personMatcher(String fieldName, matcher, Getter getter) 
    => fieldMatcher("Person", fieldName, matcher, getter);


Matcher hasStreet(matcher) => addressMatcher("street", matcher, 
    (Address a) => a.street);

Matcher addressMatcher(String fieldName, matcher, Getter getter) 
    => fieldMatcher("Address", fieldName, matcher, getter);

typedef Getter(object);

Matcher fieldMatcher(String className, String fieldName, matcher, 
                     Getter getter) =>
    new FieldMatcher(className, fieldName, matcher, getter);

class FieldMatcher extends CustomMatcher {
  final Getter getter;
  
  FieldMatcher(String className, String fieldName, matcher, this.getter) 
      : super("$className with $fieldName that", fieldName, matcher);
  
  featureValueOf(actual) => getter(actual);
}

